"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
var body_parser_1 = require("body-parser");
var compression_1 = __importDefault(require("compression"));
var cors_1 = __importDefault(require("cors"));
var express_1 = __importDefault(require("express"));
var helmet_1 = __importDefault(require("helmet"));
var __1 = require("..");
var logger_1 = require("../log/logger");
var metricsStorage_1 = require("../metrics/metricsStorage");
var serviceConfiguration_1 = require("./serviceConfiguration");
var BacktraceService = /** @class */ (function () {
    function BacktraceService(name, port, _opts, app) {
        this.name = name;
        this.port = port;
        this._opts = _opts;
        this.coronerdAddress = new Set();
        this.app = app ? app : express_1.default();
        this.configuration = BacktraceService.readConfiguration(this.name, this.port, this._opts && this._opts.confPath ? this._opts.confPath : undefined);
        var logLevel = this.configuration.logger && this.configuration.logger.logLevel
            ? this.configuration.logger.logLevel
            : undefined;
        this._logger =
            this._opts && this._opts.logger
                ? this._opts.logger
                : new logger_1.Logger(logLevel);
        if (this.configuration.backtrace &&
            this.configuration.backtrace.remoteLog !== false) {
            logger_1.Logger.initializeBacktrace(this.configuration.backtrace);
        }
        this.metrics = new metricsStorage_1.MetricsStorage(this.configuration.metrics && this.configuration.metrics.uuid, this.configuration.metrics && this.configuration.metrics.secret, undefined, this.app);
        this.prepareServiceLayer();
        BacktraceService._service = this;
    }
    /**
     * Create and startup express application with basic options.
     * By using this method, application will apply all
     * Backtrace service rules
     * @param name service name (required by service layer)
     * @param port service port (required by service layer)
     * @param opts optional application options
     * @param app  optional express app instance
     * By using this parameter you can create express app in different
     * framework and still be able to use Backtrace integrations
     */
    BacktraceService.bootstrap = function (name, port, opts, app) {
        if (!port || !name) {
            throw new Error("Invalid arguments. Name: " + name + ", port: " + port);
        }
        var service = new BacktraceService(name, port, opts, app);
        app = service.app;
        this.setupAuth(app);
        this.setupMiddlewares(app);
        this.setupServiceLayer(app, name, port, function (ip) {
            service.coronerdAddress.add(ip);
            if (opts && opts.coronerdCallback) {
                opts.coronerdCallback(ip);
            }
        });
        if (!name.startsWith('backtrace')) {
            name = "backtrace=" + name;
        }
        process.title = name;
        app.on('close', function () {
            service._logger.warn('server is closing');
        });
        return service;
    };
    BacktraceService.readConfiguration = function (name, port, path, ignoreDescriptor) {
        var serviceConfiguration = new serviceConfiguration_1.ServiceConfiguration();
        if (!path) {
            path = serviceConfiguration.getProjectSettingsPath(name);
        }
        var configuration = serviceConfiguration.loadConfiguration(path);
        if (!ignoreDescriptor) {
            configuration.descriptor = __1.getDescriptor(name, port);
        }
        return configuration;
    };
    /**
     * Setup popular Backtrace middlewares
     * @param app express app
     */
    BacktraceService.setupMiddlewares = function (app) {
        app.use(body_parser_1.json());
        app.use(body_parser_1.urlencoded({ extended: true }));
        app.use(compression_1.default());
        // Route hit by pingdom
        app.get('/', function (_req, res) {
            res.sendStatus(200);
        });
        app.use(function (err, req, res, next) {
            // handle a situation when we already sent a response from onError method
            // we shouldn't do that and we should let error middleware handle that
            // but here we wan't to make sure our code won't break up
            // and we won't generate another error
            if (res.headersSent) {
                return;
            }
            var logger = logger_1.Logger.getLogger();
            // Backtrace service ApiError (pending in CR) contains type statusCode that allows us
            // to determine what type of response we should return
            // in addition to that I recognize sometimes we retrieve status code from response.status
            //  property. This is a popular pattern for axios communication - check comments and hooks.
            if (err.statusCode || (err.response && err.response.status)) {
                var status_1 = err.statusCode || err.response.status;
                var message = err.message || err.response.message;
                if (status_1 === 500) {
                    logger.error(err);
                }
                else {
                    logger.warn(message);
                }
                res.status(status_1).send(message);
                return;
            }
            logger.error(err);
            res.sendStatus(500);
        });
    };
    /**
     * Setup security middlewares
     * @param app express app
     */
    BacktraceService.setupAuth = function (app) {
        app.use(cors_1.default());
        app.options('*', cors_1.default());
        app.use(helmet_1.default.noCache());
        app.use(helmet_1.default({
            hsts: false,
        }));
    };
    /**
     * Setup service layer integration
     * @param app express app
     */
    BacktraceService.setupServiceLayer = function (app, name, port, coronerdCallback) {
        var logger = logger_1.Logger.getLogger();
        var serviceDescriptor = this._service
            ? this._service._descriptor
            : undefined;
        // this case handle a situation when we didn't call backtraceService constr
        // here we want to read service descriptor from hard drive - if descr exists
        // we're using here a service name and port instead of service descriptor
        // to avoid extra steps required by service to setup service layer.
        if (!serviceDescriptor && name && port) {
            serviceDescriptor = __1.getDescriptor(name, port);
        }
        if (!serviceDescriptor) {
            throw new Error('Descriptor is undefined');
        }
        var identityManager = this._service
            ? this._service.identityManager
            : new __1.IdentityManager(serviceDescriptor);
        app.post("/api/" + serviceDescriptor.name + "/service", identityManager.serviceRequest({
            name: serviceDescriptor.name,
            secret: serviceDescriptor.secret,
            coronerdCallback: coronerdCallback,
        }));
    };
    BacktraceService.prototype.start = function () {
        var _this = this;
        if (!this.app) {
            this._logger.error("Express application is not defined. Cannot start application");
            return;
        }
        if (!this._descriptor) {
            this._logger.error("Descriptor object is not defined.\n        Please execute setupServiceLayer method");
            return;
        }
        __1.listenDescriptor(this._descriptor, this.app, this.configuration.ssl, function () {
            _this._logger.info("Application is Running. " + (_this._descriptor ? "Port: " + _this._descriptor.port : ''));
        });
    };
    BacktraceService.prototype.prepareServiceLayer = function () {
        if (!this._descriptor) {
            var descriptor = __1.getDescriptor(this.name, this.port);
            this._descriptor = descriptor;
        }
        this.identityManager = new __1.IdentityManager(this._descriptor);
    };
    return BacktraceService;
}());
exports.BacktraceService = BacktraceService;
//# sourceMappingURL=backtraceService.js.map