import { randomBytes } from 'crypto';
import * as fs from 'fs';
import * as http from 'http';
import * as https from 'https';
import mkdirp from 'mkdirp';
import path from 'path';
import * as url from 'url';
import { IServiceDescriptor } from '../model/serviceDescriptor';
import { ICoronerConfig } from './model/coronerConfig';
import { ICoronerDescriptor } from './model/coronerDescriptor';
import { IDescriptorOpts } from './model/descriptorOptions';

// XXX: better default?
const defaultCoronerdHostname = '0.0.0.0';

function readJSONFile(filePath: string): any | undefined {
  let obj = null;
  try {
    obj = JSON.parse(fs.readFileSync(filePath, { encoding: 'utf8' }));
  } catch (e) {
    return undefined;
  }
  return obj;
}

function getCoronerdHostname(crcfg: ICoronerConfig): string {
  const alldescPath =
    process.env.ALL_DESCRIPTOR_PATH || `${crcfg.varprefix}/_all.json`;
  // if coronerd has supplied its REST URL, then use it.
  if (!fs.existsSync(alldescPath)) {
    return defaultCoronerdHostname;
  }
  const alldesc = readJSONFile(alldescPath);
  if (alldesc && typeof alldesc.coronerd === 'string') {
    const crUrl = url.parse(alldesc.coronerd);
    if (crUrl.hostname) {
      return crUrl.hostname;
    }
  }
  return defaultCoronerdHostname;
}

export function listenDescriptor(
  descr: IServiceDescriptor,
  app: http.RequestListener,
  credentials: https.ServerOptions | undefined,
  callback: () => void | undefined,
): void {
  let server;
  switch (descr.protocol) {
    case 'http':
      server = http.createServer(app);
      break;
    case 'https':
      if (!credentials) {
        throw new Error(`Empty credentials.`);
      }
      server = https.createServer(credentials, app);
      break;
    default:
      throw new Error(`unknown protocol ${descr.protocol}`);
  }
  server.listen(descr.port, callback);
}

export function getCoronerConfig(): ICoronerConfig {
  const cfg: ICoronerConfig = {
    svcDescriptorPrefix: process.env.SERVICE_DESCRIPTOR_VARPREFIX,
    svclayerPrefix: process.env.SERVICE_LAYER_VARPREFIX,
    filePath: process.env.CORONERD_CONF || '/etc/coronerd/coronerd.conf',
    varprefix: undefined,
    json: undefined,
  };

  cfg.json = readJSONFile(cfg.filePath);
  if (cfg.json) {
    if (cfg.json.svclayer) {
      const descriptionExists =
        Array.isArray(cfg.json.svclayer.descriptorpaths) &&
        cfg.json.svclayer.descriptorpaths.length > 0;
      if (descriptionExists) {
        /*
         * Prefer the last element as it will customarily be one that the
         * service can write new files to.
         */
        const paths = cfg.json.svclayer.descriptorpaths;
        cfg.svcDescriptorPrefix = paths[paths.length - 1];
      }
      if (!cfg.svclayerPrefix && cfg.json.svclayer.varpath) {
        cfg.svclayerPrefix = cfg.json.svclayer.varpath;
      }
    }
    cfg.varprefix = cfg.json.varprefix;
  }

  /* Implement hardcoded fallbacks that match coronerd's. */
  if (!cfg.varprefix) {
    cfg.varprefix = '/var/run/coronerd';
  }
  if (!cfg.svclayerPrefix) {
    cfg.svclayerPrefix = `${cfg.varprefix}/svclayer`;
  }
  if (!cfg.svcDescriptorPrefix) {
    cfg.svcDescriptorPrefix = `${cfg.varprefix}/services.d`;
  }
  return cfg;
}

/**
 * getDescriptor fetches the service's integration parameters.
 * @param serviceName - service name - for example: share/saml/comments
 * @param defaultPort - the default port number for the service.
 * @param opts - descriptor options
 */
export function getDescriptor(
  serviceName: string,
  defaultPort: number,
  opts: IDescriptorOpts = {},
): IServiceDescriptor {
  const crcfg = getCoronerConfig();
  const varPath =
    process.env.SERVICE_DESCRIPTOR_VARPATH ||
    `${crcfg.svcDescriptorPrefix}/${serviceName}.json`;

  const paths: string[] = getDescriptorPaths(serviceName, varPath);
  let foundPath = '';
  let desc: ICoronerDescriptor | undefined;
  for (const descriptorPath of paths) {
    desc = readDescriptorConfiguration(descriptorPath);
    if (desc) {
      foundPath = descriptorPath;
      break;
    }
  }

  // if descriptor doesn't exist in the descriptor search path,
  // generate new descriptor file with default configuration in the
  // var path, where it can own the file.
  if (!desc) {
    desc = getDefaultDescriptor(serviceName, defaultPort, crcfg, opts);
    mkdirp(path.dirname(varPath), (err: Error) => {
      if (err) {
        throw err;
      }
      fs.writeFileSync(varPath, JSON.stringify(desc, null, 2));
    });
  }
  const surl = new url.URL(desc.server);

  // if the descriptor owned by the service was read in and happens to
  // contain the invalid-for-SSL default hostname "0.0.0.0" that was
  // written out previously, see whether it can be automatically fixed up.
  if (foundPath === varPath && surl.hostname === defaultCoronerdHostname) {
    const hostname = getCoronerdHostname(crcfg);
    if (hostname !== defaultCoronerdHostname) {
      surl.hostname = hostname;
      desc.server = surl.toString();
      fs.writeFileSync(varPath, JSON.stringify(desc, null, 2));
    }
  }

  let port: number = surl.protocol === 'https:' ? 443 : 80;
  if (surl.port) {
    port = parseInt(surl.port, 10);
  }
  return {
    name: desc.name,
    port,
    resource: desc.resource,
    secret: desc.secret,
    isOnPremise: desc.proxy,
    protocol: surl.protocol.split(':')[0],
  };
}

function getDefaultDescriptor(
  serviceName: string,
  defaultPort: number,
  crcfg: ICoronerConfig,
  opts: IDescriptorOpts,
): ICoronerDescriptor {
  const protocol = opts.protocol ? opts.protocol : 'http';
  const hostname = getCoronerdHostname(crcfg);
  return {
    name: serviceName,
    server: opts.server || `${protocol}://${hostname}:${defaultPort}`,
    secret: randomBytes(32).toString('hex'),
    resource: opts.resource || `/api/${serviceName}`,
    proxy: opts.proxy !== undefined ? opts.proxy : true,
    enabled: true,
  };
}

function readDescriptorConfiguration(
  descriptorPath: string,
): ICoronerDescriptor | undefined {
  return readJSONFile(descriptorPath);
}

function getDescriptorPaths(serviceName: string, varPath: string): string[] {
  return !process.env.SERVICE_DESCRIPTOR_PATH
    ? [`/etc/coronerd/services.d/${serviceName}.json`, varPath]
    : [process.env.SERVICE_DESCRIPTOR_PATH];
}
