import axios from 'axios';
import express, { NextFunction, Request, Response } from 'express';
import os from 'os';
export class MetricsStorage {
  public beforeSend?: () => Promise<void>;

  private readonly _hostUrl: string =
    'https://api.circonus.com/module/httptrap/';
  private _url = `${this._hostUrl}${this.uuid}/${this.secret}`;

  private readonly _hostName: string = os.hostname().split('.')[0];

  private _data: {
    [key: string]: {
      _type: string;
      _value: number[];
    };
  } = {};

  constructor(
    private readonly uuid: string,
    private readonly secret: string,
    private opts?: { interval?: number; hostUrl?: string },
    app?: express.Express,
    private logger?: { warn: (msg: string) => void | Promise<void> },
  ) {
    if (!this.uuid || !this.secret) {
      if (logger) {
        logger.warn(
          'Opts (or uuid/secret) object is required. Metrics are disabled',
        );
      } else {
        console.warn(
          'Opts (or uuid/secret) object is required. Metrics are disabled',
        );
      }
      return;
    }

    // override default circonus url.
    // different url might be useful in case
    // when we will prepare our own metrics client
    if (this.opts && this.opts.hostUrl) {
      this._hostUrl = this.opts.hostUrl;
    }

    if (app) {
      app.use((req, res, next) => this.registerRequestLatency(req, res, next));
    }
    this.addMetric('birthday', Date.now());

    setInterval(async () => {
      if (this.beforeSend) {
        await this.beforeSend();
      }
      if (Object.keys(this._data).length > 0) {
        this.send();
      }
    }, (this.opts && this.opts.interval) || 1000);
  }

  /**
   * Convert objects to metrics
   * @param source metrics object
   */
  public add(source: object) {
    // metrics are disabled because of missing secret and uuid parameters
    if (!this.uuid || !this.secret) {
      return;
    }
    for (const key in source) {
      if (source.hasOwnProperty(key)) {
        const value = (source as any)[key];
        this.addMetric(key, value);
      }
    }
  }

  /**
   * Add single value to metrics
   * @param name metrics name
   * @param value metrics value
   */
  public addMetric(name: string, value: number) {
    // metrics are disabled because of missing secret and uuid parameters
    if (!this.uuid || !this.secret) {
      return;
    }
    name = `${name}|ST[hostname:${this._hostName}]`;
    this._data[name] = this._data[name] || { _type: 'n', _value: [] };
    this._data[name]._value.push(value);
  }

  private registerRequestLatency(
    req: Request,
    res: Response,
    next: NextFunction,
  ): void {
    if (!req.path) {
      return;
    }

    const startHrTime = process.hrtime();
    res.on('finish', () => {
      const elapsedHrTime = process.hrtime(startHrTime);
      const elapsedTimeInMs = elapsedHrTime[0] * 1000 + elapsedHrTime[1] / 1e6;
      this.addMetric(
        'express`' + req.method + '`' + req.path + '`latency',
        elapsedTimeInMs,
      );
    });
    next();
  }

  private async send(): Promise<void> {
    const data = this._data;
    this._data = {};
    try {
      await axios.put(this._url, data);
    } catch (error) {
      if (this.logger) {
        this.logger.warn(`Cannot send metrics. Reason: ${error.message}`);
      }
    }
  }
}
