const express = require('express');
const passport = require('passport');
const MultiSamlStrategy = require('passport-saml/multiSamlStrategy');
const session = require('express-session');
const FileStore = require('session-file-store')(session);
const secureRandomString = require('secure-random-string');
const cookieParser = require('cookie-parser');
const { BacktraceService } = require('backtrace-service');
const { service } = require('./util/service-layer');
const getSamlOptions = require('./util/getSamlOptions');
const { authenticateCoronerd } = require('./util/coronerdTokenCache');

const config = service.configuration;

const PATH_PREFIX = config.pathPrefix;

// exportable function for creating express app to be executed in src/index.js
// allows us to export entire express app for integration testing :)
module.exports = () => {
  const samlStrategy = new MultiSamlStrategy(
    {
      passReqToCallback: true,
      getSamlOptions: getSamlOptions,
    },
    function (req, profile, done) {
      return done(null, profile);
    }
  );

  const app = service.app;

  BacktraceService.setupAuth(app);

  BacktraceService.setupMiddlewares(app);

  BacktraceService.setupServiceLayer(app, service.name, service.port, (url) => {
    service._logger.info(`Service heard from coronerd at ${url}`);
    authenticateCoronerd(url);
  });

  app.set('view engine', 'ejs');

  app.set('views', __dirname + '/views');

  app.use(express.static(__dirname + '/public'));

  app.use(cookieParser());

  if (process.env.NODE_ENV === 'production') {
    app.use(
      session({
        secret: secureRandomString({ length: 64, alphanumeric: true }),
        store: new FileStore({ ttl: 600 }),
        resave: true,
        saveUninitialized: true,
      })
    );
  } else {
    app.use(
      session({
        secret: secureRandomString({ length: 64, alphanumeric: true }),
        resave: true,
        saveUninitialized: true,
      })
    );
  }

  passport.use(samlStrategy);

  app.use(passport.initialize());

  app.use(passport.session());

  passport.serializeUser(function (user, done) {
    done(null, user);
  });

  passport.deserializeUser(function (user, done) {
    done(null, user);
  });

  // redundant routes for legacy will eventually be able to get rid of this
  app.use('/', require('./routes')(passport));

  // service layer prefixes routes
  app.use(PATH_PREFIX, require('./routes')(passport));

  return app;
};
