#!/usr/bin/env node

const cluster = require('cluster');
const http = require('http');
const https = require('https');
const sendReport = require('./util/sendReport');
const { service, ports } = require('./util/service-layer');
const createApp = require('./createApp');

const { configuration } = service;

// If master then fork workers Else start server as worker
if (cluster.isMaster) {
  // get number of cpu cores available

  service._logger.info('Master cluster is starting up worker...');

  // fork a worker
  cluster.fork();

  // listener for worker as they start up
  cluster.on('online', (worker) => {
    service._logger.info(`Worker ${worker.process.pid} is online`);
  });

  // listener for worker as they stop and then handling restart
  cluster.on('exit', (worker, code, signal) => {
    service._logger.warn(
      `Worker ${worker.process.pid} died with a code:\n${code}\nand signal:${signal}\nStarting new worker`
    );
    cluster.fork();
  });
} else {
  if (process.listeners('uncaughtException').length == 0) {
    process.on('uncaughtException', function (error) {
      service._logger.error(error + '');
      sendReport(error);
      process.exit(1);
    });
  }

  // TODO remove this once all hosts are on service layer with proper descriptors
  const getProtocol = (service) => {
    if (service._descriptor && service._descriptor.protocol) {
      return service._descriptor.protocol;
    }
    return configuration && configuration.ssl ? 'https' : 'http';
  };

  // Port and protocol have to be passed explicitly because we have some hosted
  // versions of this application that are listening on many ports and not just
  // on the ports specified in the service descriptor
  const startServer = (port, protocol, credentials) => {
    const app = createApp();
    let server;

    switch (protocol) {
      case 'http':
        server = http.createServer(app);
        break;
      case 'https':
        server = https.createServer(credentials, app);
        break;
      default:
        throw new Error(`unknown protocol ${protocol}`);
    }

    server.listen(port, () =>
      service._logger.info(`Server is listening on port ${port}`)
    );
  };

  // Start service
  ports.forEach((port) => {
    startServer(port, getProtocol(service), configuration.ssl);
  });
}
