const express = require('express');
const sendReport = require('../util/sendReport');
const getVersion = require('../util/getVersion');
const Tenant = require('../models/tenant');
const axios = require('axios');
const { service } = require('../util/service-layer');

const APP_VERSION = getVersion();

// Abstract access rights for easy update
const hasWriteAccess = (user) => {
  return user.role === 'admin';
};

const prepAuth = (req, res, next) => {
  let auth;
  // conditional for Legacy systems not expecting headers
  // TODO remove once all FE requests utilize X-Coroner-* headers in all API versions
  if (req.get('X-Coroner-Token') && req.get('X-Coroner-Location')) {
    auth = {
      token: req.get('X-Coroner-Token'),
      url: req.get('X-Coroner-Location'),
    };
  } else {
    auth = {
      token: req.query.token,
      url: 'https://' + req.query.host,
    };
  }

  req.coronerAuth = auth;

  next();
};

// TODO remove once saml.backtrace.io is on SL
const authRequest = (req, res, next) => {
  const { token, url } = req.coronerAuth;
  axios
    .get(`${url}/api/config`, {
      headers: {
        'X-Coroner-Token': token,
      },
    })
    .then((result) => {
      if (result && result.data) {
        req.coronerAuth.cfg = result.data;
        next();
      } else {
        next(new Error('Failed to authenticate'));
      }
    })
    .catch(next);
};

const authenticateRequest = (req, res, next) => {
  const { configuration } = service;
  if (configuration && configuration.serviceLayerDisabled) {
    authRequest(req, res, next);
  } else {
    service.identityManager.authenticateRequest()(req, res, next);
  }
};

const getTenantAction = (action, form) => {
  switch (action) {
    case 'modify':
      return Tenant.modify(form);
    case 'create':
      return Tenant.create(form);
    case 'delete':
      return Tenant.delete(form.hostname);
    default:
      throw new Error('invalid action');
  }
};

module.exports = (passport) => {
  const router = express.Router();

  // Route for retreiving tenant configuration from db
  // used by console 2.0's UI for SSO config
  router.get('/', prepAuth, authenticateRequest, (req, res, next) => {
    const start = Date.now();
    Tenant.get(req.query.host)
      .then((data) => {
        res.status(200).json({
          response: { tenant: data[0] || {} },
          _: {
            version: APP_VERSION,
            latency: Date.now() - start,
          },
        });
      })
      .catch((error) => {
        sendReport(error, { hostname: req.tenant });
        next(new Error('Failed to retrieve tenant data'));
      });
  });

  // Route for creating/modifying SAML config for tenant
  // used by console 2.0's UI for SSO config
  router.post('/', prepAuth, authenticateRequest, (req, res, next) => {
    try {
      const user = req.coronerAuth.cfg.user;
      if (hasWriteAccess(user)) {
        const start = Date.now();
        if (req.body && req.body.action) {
          getTenantAction(req.body.action, req.body.form)
            .then(() => {
              // if (req.body.action == 'modify' || req.body.action == 'delete') {
              //   passports.clearInstance(req.body.form.hostname);
              // }
              res.status(200).json({
                response: {
                  status: 'ok',
                },
                _: {
                  version: APP_VERSION,
                  latency: Date.now() - start,
                },
              });
            })
            .catch((error) => {
              sendReport(error, { hostname: req.tenant });
              next(new Error(`failure to ${req.body.action} tenant`));
            });
        } else {
          next(new Error('malformed request'));
        }
      } else {
        next(new Error('access denied'));
      }
    } catch (err) {
      next(err);
    }
  });

  // Route for allowing the client to query service if service for tenant exists
  // ex: GET https://saml.backtrace.io/test?tenant=universe.sp.backtrace.io
  router.get('/test', (req, res, next) => {
    const start = Date.now();
    Tenant.get(req.query.tenant)
      .then((data) =>
        data.length > 0
          ? res.json({
              response: { status: 'ok' },
              _: {
                version: APP_VERSION,
                latency: Date.now() - start,
              },
            })
          : next(new Error('Failed to retrieve tenant data'))
      )
      .catch((error) => {
        sendReport(error, { hostname: req.tenant });
        next(new Error('Failed to retrieve tenant data'));
      });
  });

  router.use('/*', (error, req, res, next) => {
    res.status(400).json({
      error: error.name,
      message: error.message,
    });
  });

  return router;
};
